﻿
using Silk.NET.OpenGL;
using StbImageSharp;

namespace Engine_silk.NET.Textures;

public class Texture2D : IDisposable
{
	/// <summary>
	/// Defines how the texture should be wrapped if the uv-coordinates are out of bounds
	/// Default: Repeat
	/// </summary>
	public TextureWrapMode WrapMode { get; set; }
	/// <summary>
	/// Defines how a texel should be interpolated when zooming out of the texture.
	/// Default: BiLinear
	/// </summary>
	public TextureMinFilter MinifyingInterpolation { get; set; }
	/// <summary>
	/// Defines how a texel should be interpolated when zooming in to the texture.
	/// Default: Bilinear
	/// </summary>
	public TextureMagFilter MagnifyingInterpolation { get; set; }

	protected GL _glContext;
	protected uint _textureId;
	private bool disposedValue;

	public Texture2D(GL glContext, string pathToTexture, ImageFormat format = ImageFormat.RGB, TextureWrapMode wrapMode = TextureWrapMode.Repeat,
		TextureMinFilter minifyingInterpolation = TextureMinFilter.Linear, TextureMagFilter magnifyingInterpolation = TextureMagFilter.Linear)
	{
		_glContext = glContext;
		WrapMode = wrapMode;
		MinifyingInterpolation = minifyingInterpolation;
		MagnifyingInterpolation = magnifyingInterpolation;

		byte[] buffer = File.ReadAllBytes(pathToTexture);
		ImageResult image = ImageResult.FromMemory(buffer, (ColorComponents)format);
		ReadOnlySpan<byte> imageData = new(image.Data);

		_glContext.GenTextures(1, out _textureId);
		_glContext.BindTexture(TextureTarget.Texture2D, _textureId);
		_glContext.TexImage2D(TextureTarget.Texture2D, 0, InternalFormat.Rgb, (uint)image.Width, (uint)image.Height, 0, format.ImageFormatToPixelFormat(), PixelType.UnsignedByte, imageData);
		_glContext.GenerateMipmap(TextureTarget.Texture2D);
	}

	public void Bind(TextureUnit textureUnit = TextureUnit.Texture0)
	{
		_glContext.TexParameter(TextureTarget.Texture2D, GLEnum.TextureWrapT, (int)WrapMode);
		_glContext.TexParameter(TextureTarget.Texture2D, GLEnum.TextureMinFilter, (int)MinifyingInterpolation);
		_glContext.TexParameter(TextureTarget.Texture2D, GLEnum.TextureMagFilter, (int)MagnifyingInterpolation);

		_glContext.ActiveTexture(textureUnit);
		_glContext.BindTexture(TextureTarget.Texture2D, _textureId);
	}

	protected virtual void Dispose(bool disposing)
	{
		if (!disposedValue)
		{
			if (disposing)
			{
				_glContext.DeleteTexture(_textureId);
			}

			// TODO: free unmanaged resources (unmanaged objects) and override finalizer
			// TODO: set large fields to null
			disposedValue = true;
		}
	}

	// // TODO: override finalizer only if 'Dispose(bool disposing)' has code to free unmanaged resources
	// ~Texture2D()
	// {
	//     // Do not change this code. Put cleanup code in 'Dispose(bool disposing)' method
	//     Dispose(disposing: false);
	// }

	public void Dispose()
	{
		// Do not change this code. Put cleanup code in 'Dispose(bool disposing)' method
		Dispose(disposing: true);
		GC.SuppressFinalize(this);
	}
}
